" Generate landscape in the style of https://lodev.org/cgtutor/randomnoise.html "
import random

BOX_CHAR = '▀'
ESC = chr(27)
GREYS = [0] + list(range(232, 256)) + [231]  # Darkest to lightest
INKY = [0, 124, 231]
FULLCOLOUR = [17, 108, 130, 22, 142]

def colour(fg, bg):
    return f'{ESC}[38;5;{fg}m{ESC}[48;5;{bg}m'

def reset():
    return f'{ESC}[0m'

def output_as_text(arr, width, colours):
    """
    Output 'arr' as 256 shades of grey.

    Arr is expected to be floats in the range 0-1 which are converted to one of the 26 shades of grey (including white
    and black) in xterm-256color.
    """
    grey = lambda f: colours[int(f * (len(colours) - 1))]
    clamp = lambda x: min(max(x, 0), 1)

    idx = 0
    line = []
    last_idx = len(arr) - width
    while idx < last_idx:
        top = clamp(arr[idx])
        bottom = clamp(arr[idx + width])
        line.append(colour(grey(top), grey(bottom)) + BOX_CHAR)

        idx += 1
        if (idx % width) == 0:
            print(''.join(line) + reset())
            line = []
            idx += width

def raw_noise(width, height):
    return [ [random.random() for _x in range(width)] for _y in range(height)]

def bilinear(tex, x, y):
    y = y * len(tex)
    x = x * len(tex[0])

    fract_x = x - int(x)
    fract_y = y - int(y)

    y1 = int(y)
    x1 = int(x)

    y2 = (y1 - 1) % len(tex)
    x2 = (x1 - 1) % len(tex[0])

    return (
        (fract_x * fract_y * tex[y1][x1])
        + ((1 - fract_x) * fract_y * tex[y1][x2])
        + (fract_x * (1 - fract_y) * tex[y2][x1])
        + ((1 - fract_x) * (1 - fract_y) * tex[y2][x2]))

def turbulence(tex, x, y, size):
    initial_size = size
    value = 0

    while size >= 1:
        value += (bilinear(tex, x / size, y / size) * size)
        size /= 2

    value /= (initial_size * 2)

    return value

def nonlinear_3col(f):
    if f > 0.6:
        return 1
    elif f > 0.56:
        return 0.7
    return 0

def nonlinear_4col(f):
    if f > 0.6:
        return 1
    elif f > 0.57:
        return 0.7
    elif f > 0.54:
        return 0.3
    return 0

def nonlinear_5col(f):
    if f > 0.73:
        return 1
    elif f > 0.6:
        return 0.75
    elif f > 0.57:
        return 0.6
    elif f > 0.54:
        return 0.4
    return 0

def nonlinear_none(f):
    return f

def random_landscape(slice_width, slice_height, nonlinearity=nonlinear_3col):
    out_width = 256
    out_height = 256

    #noise = raw_noise(int(out_width / 2), int(out_height / 2))
    noise = raw_noise(out_width, out_height)

    #magnified = [bilinear(noise, (x / out_width) / 8, (y / out_height) / 8)
    #             for x in range(out_width)
    #             for y in range(out_height)]
    turbulent = [
        [turbulence(noise, (x / out_width), (y / out_height), 32) for x in range(out_width)]
        for y in range(out_height)]

    sliced_1d = []
    for y in range(slice_height):
        sliced_1d.extend(turbulent[y][:slice_width])

    filtered = [nonlinearity(f) for f in sliced_1d]

    return filtered

def output_random_landscape():
    slice_width = 212
    slice_height = 104

    output_as_text(random_landscape(slice_width, slice_height), slice_width, INKY)


if __name__ == '__main__':
    output_random_landscape()

